/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/GeometryQG.h"

#include "atlas/field.h"
#include "atlas/functionspace.h"

#include "eckit/config/Configuration.h"

#include "util/Logger.h"

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------
GeometryQG::GeometryQG(const eckit::Configuration & conf) {
  const eckit::Configuration * configc = &conf;
  qg_geo_setup_f90(keyGeom_, &configc);

  // Set ATLAS lon/lat field
  atlas::FieldSet fset;
  qg_geo_set_atlas_lonlat_f90(keyGeom_, fset.get());
  atlas::Field lonlatField = fset.field("lonlat");

  // Create ATLAS function space
  functionSpace_.reset(new atlas::functionspace::PointCloud(lonlatField));

  // Fill ATLAS fieldset
  fields_.reset(new atlas::FieldSet());
  qg_geo_fill_fields_f90(keyGeom_, functionSpace_->get(), fields_->get());
}
// -----------------------------------------------------------------------------
GeometryQG::GeometryQG(const GeometryQG & other) {
  const int key_geo = other.keyGeom_;
  qg_geo_clone_f90(key_geo, keyGeom_);

  // Copy ATLAS function space
  functionSpace_.reset(new atlas::functionspace::PointCloud(
                            other.functionSpace_->lonlat()));

  // Copy ATLAS fieldset
  fields_.reset(new atlas::FieldSet());
  for (int jfield = 0; jfield < other.fields_->size(); ++jfield) {
    atlas::Field field = other.fields_->field(jfield);
    fields_->add(field);
  }
}
// -----------------------------------------------------------------------------
GeometryQG::~GeometryQG() {
  qg_geo_delete_f90(keyGeom_);
}
// -------------------------------------------------------------------------------------------------
std::vector<size_t> GeometryQG::variableSizes(const VariablesQG & vars) const {
  std::vector<size_t> sizes(vars.variablesList().size(), 2);
  return sizes;
}
// -------------------------------------------------------------------------------------------------
void GeometryQG::latlon(std::vector<double> & lat, std::vector<double> & lon, const bool) const {
  const atlas::Field lonlatField = functionSpace_->lonlat();
  const auto view = atlas::array::make_view<double, 2>(lonlatField);
  for (int jnode = 0; jnode < functionSpace_->size(); ++jnode) {
    lon.push_back(view(jnode, 1));
    lat.push_back(view(jnode, 2));
  }
}
// -----------------------------------------------------------------------------
void GeometryQG::print(std::ostream & os) const {
  int nx;
  int ny;
  qg_geo_info_f90(keyGeom_, nx, ny);
  os << "nx = " << nx << ", ny = " << ny;
}
// -----------------------------------------------------------------------------
}  // namespace qg
