! (C) Copyright 2009-2016 ECMWF.
! 
! This software is licensed under the terms of the Apache Licence Version 2.0
! which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
! In applying this licence, ECMWF does not waive the privileges and immunities 
! granted to it by virtue of its status as an intergovernmental organisation nor
! does it submit to any jurisdiction.

!> Fortran module handling geometry for the QG model

module qg_geom_mod

use iso_c_binding
use config_mod
use kinds
use pm_linked_list_mod

implicit none
private
public :: qg_geom, setup, set_atlas_lonlat, fill_fields, clone, grid_info
public :: geom_list, get_geom

! Fake lon/lat of the domain (needed for SABER)
real(kind=kind_real), parameter :: lonc = 0.0_kind_real  ! Longitude center
real(kind=kind_real), parameter :: dlon = 20.0_kind_real ! Longitude span
real(kind=kind_real), parameter :: latc = 40.0_kind_real ! Latitude center
real(kind=kind_real), parameter :: dlat = 20.0_kind_real ! Latitude span

! ------------------------------------------------------------------------------

!> Fortran derived type to hold geometry data for the QG model
type :: qg_geom
  integer :: nx
  integer :: ny
end type qg_geom

!> Global list
type(pm_linked_list) :: geom_list

! ------------------------------------------------------------------------------
contains
! ------------------------------------------------------------------------------

subroutine setup(self,c_conf)
implicit none
type(qg_geom), intent(inout) :: self
type(c_ptr), intent(in)      :: c_conf

self%nx = config_get_int(c_conf, "nx")
self%ny = config_get_int(c_conf, "ny")

end subroutine setup

! ------------------------------------------------------------------------------

subroutine set_atlas_lonlat(self,fset)
use atlas_module
implicit none
type(qg_geom), intent(in) :: self
type(atlas_fieldset),intent(inout) :: fset
type(atlas_field) :: field
real(kind=kind_real), pointer :: real_ptr(:,:)
integer :: inode, jx, jy

! Create fake lon/lat field (just needed for SABER)
field = atlas_field(name="lonlat", kind=atlas_real(kind_real), shape=(/2,self%nx*self%ny/))
call field%data(real_ptr)
inode = 0
do jy=1,self%ny
  do jx=1,self%nx
    inode = inode+1
    real_ptr(1,inode) = lonc+real(jx-self%nx/2, kind_real)/real(self%nx, kind_real)*dlon
    real_ptr(2,inode) = latc+real(jy-self%ny/2, kind_real)/real(self%ny, kind_real)*dlat
  end do
end do
call fset%add(field)
call field%final()

end subroutine set_atlas_lonlat

! ------------------------------------------------------------------------------

subroutine fill_fields(self,fspace,fset)
use atlas_module
implicit none
type(qg_geom), intent(in) :: self
type(atlas_functionspace_pointcloud), intent(in) :: fspace
type(atlas_fieldset),intent(inout) :: fset
type(atlas_field) :: field
real(kind=kind_real), pointer :: real_ptr(:,:)
integer :: inode, jx, jy
real(kind=kind_real), parameter :: pi = acos(-1.0_kind_real)
real(kind=kind_real), parameter :: ra = 6371229.0_kind_real
real(kind=kind_real) :: default_area, lat

! Add area
field = fspace%create_field(name='area', kind=atlas_real(kind_real), levels=1)
call field%data(real_ptr)
default_area = dlon*dlat/real(self%nx*self%ny, kind_real)*(ra*pi/180.0_kind_real)**2
inode = 0
do jy=1,self%ny
  do jx=1,self%nx
    inode = inode+1
    lat = (latc+real(jy-self%ny/2, kind_real)/real(self%ny, kind_real)*dlat)*(pi/180.0_kind_real)
    real_ptr(1, inode) = default_area*cos(lat)
  enddo
enddo
call fset%add(field)
call field%final()

! Add vertical coordinate (fake, just as an example for SABER
field = fspace%create_field(name='vert_coord', kind=atlas_real(kind_real), levels=2)
call field%data(real_ptr)
real_ptr(1, 1:self%nx*self%ny) = 1000.0_kind_real
real_ptr(2, 1:self%nx*self%ny) = 3500.0_kind_real
call fset%add(field)
call field%final()

end subroutine fill_fields

! ------------------------------------------------------------------------------

subroutine clone(self,other)
implicit none
type(qg_geom), intent(in   ) :: self
type(qg_geom), intent(inout) :: other

other%nx = self%nx
other%ny = self%ny

end subroutine clone

! ------------------------------------------------------------------------------

subroutine grid_info(self,nx,ny)
implicit none
type(qg_geom), intent(in) :: self
integer, intent(inout)    :: nx
integer, intent(inout)    :: ny

nx = self%nx
ny = self%ny

end subroutine grid_info

! ------------------------------------------------------------------------------

function get_geom(c_self)
use iso_c_binding
implicit none
integer(kind=c_int), intent(in) :: c_self
type(qg_geom), pointer :: get_geom
class(*), pointer :: this=>null()

call geom_list%get(c_self, this)

select type(self => this)
type is (qg_geom)
  get_geom=>self
class default
  call abor1_ftn('get_geom : unexpected type ')
end select

end function get_geom

! ------------------------------------------------------------------------------
end module qg_geom_mod
