/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#include "oops/generic/GenericCtlVec.h"

#include <memory>
#include <random>
#include <string>

#include "atlas/array.h"
#include "atlas/field.h"
#include "eckit/exception/Exceptions.h"
#include "util/Logger.h"
#include "util/Mpi.h"
#include "util/Timer.h"
#include "util/abor1_cpp.h"

namespace oops {
// =============================================================================
/// Constructors, destructor
// -----------------------------------------------------------------------------

GenericCtlVec::GenericCtlVec(const size_t &n) {
  Log::trace() << "GenericCtlVec::GenericCtlVec starting" << std::endl;
  util::Timer timer(classname(), "GenericCtlVec");
  data_ = atlas::Field("genericCtlVec", atlas::array::make_datatype<double>(),
                       atlas::array::make_shape(n));
  Log::trace() << "GenericCtlVec::GenericCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

GenericCtlVec::GenericCtlVec(const size_t &n, const GenericCtlVec &other) {
  Log::trace() << "GenericCtlVec::GenericCtlVec starting" << std::endl;
  util::Timer timer(classname(), "GenericCtlVec");
  data_ = atlas::Field("genericCtlVec", atlas::array::make_datatype<double>(),
                       atlas::array::make_shape(n));
  ASSERT(n == other.size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewOther = atlas::array::make_view<double, 1>(other.data_);
  for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
    view(jnode) = viewOther(jnode);
  }
  Log::trace() << "GenericCtlVec::GenericCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

GenericCtlVec::GenericCtlVec(const GenericCtlVec &other, const bool copy) {
  Log::trace() << "GenericCtlVec::GenericCtlVec copy starting" << std::endl;
  util::Timer timer(classname(), "GenericCtlVec");
  data_ = atlas::Field("genericCtlVec", atlas::array::make_datatype<double>(),
                       atlas::array::make_shape(other.size()));
  if (copy) {
    auto view = atlas::array::make_view<double, 1>(data_);
    auto viewOther = atlas::array::make_view<double, 1>(other.data_);
    for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
      view(jnode) = viewOther(jnode);
    }
  }
  Log::trace() << "GenericCtlVec::GenericCtlVec copy done" << std::endl;
}

// -----------------------------------------------------------------------------

GenericCtlVec::~GenericCtlVec() {
  Log::trace() << "GenericCtlVec::~GenericCtlVec starting" << std::endl;
  util::Timer timer(classname(), "~Incr");
  Log::trace() << "GenericCtlVec::~GenericCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::zero() {
  Log::trace() << "GenericCtlVec::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  auto view = atlas::array::make_view<double, 1>(data_);
  for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
    view(jnode) = 0.0;
  }
  Log::trace() << "GenericCtlVec::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

GenericCtlVec &GenericCtlVec::operator=(const GenericCtlVec &rhs) {
  Log::trace() << "GenericCtlVec::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  ASSERT(rhs.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewRhs = atlas::array::make_view<double, 1>(rhs.data_);
  for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
    view(jnode) = viewRhs(jnode);
  }
  Log::trace() << "GenericCtlVec::operator= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

GenericCtlVec &GenericCtlVec::operator+=(const GenericCtlVec &rhs) {
  Log::trace() << "GenericCtlVec::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");
  ASSERT(rhs.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewRhs = atlas::array::make_view<double, 1>(rhs.data_);
  for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
    view(jnode) += viewRhs(jnode);
  }
  Log::trace() << "GenericCtlVec::operator+= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

GenericCtlVec &GenericCtlVec::operator-=(const GenericCtlVec &rhs) {
  Log::trace() << "GenericCtlVec::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");
  ASSERT(rhs.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewRhs = atlas::array::make_view<double, 1>(rhs.data_);
  for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
    view(jnode) -= viewRhs(jnode);
  }
  Log::trace() << "GenericCtlVec::operator-= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

GenericCtlVec &GenericCtlVec::operator*=(const double &zz) {
  Log::trace() << "GenericCtlVec::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");
  auto view = atlas::array::make_view<double, 1>(data_);
  for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
    view(jnode) *= zz;
  }
  Log::trace() << "GenericCtlVec::operator*= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::axpy(const double &zz, const GenericCtlVec &dx,
                         const bool check) {
  Log::trace() << "GenericCtlVec::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");
  ASSERT(dx.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewDx = atlas::array::make_view<double, 1>(dx.data_);
  for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
    view(jnode) += zz * viewDx(jnode);
  }
  Log::trace() << "GenericCtlVec::axpy done" << std::endl;
}

// -----------------------------------------------------------------------------

double GenericCtlVec::local_dot_product_with(const GenericCtlVec &dx) const {
  Log::trace() << "GenericCtlVec::local_dot_product_with starting" << std::endl;
  util::Timer timer(classname(), "local_dot_product_with");
  double zz = 0;
  ASSERT(dx.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewDx = atlas::array::make_view<double, 1>(dx.data_);
  for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
    zz += view(jnode) * viewDx(jnode);
  }
  Log::trace() << "GenericCtlVec::local_dot_product_with done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::schur_product_with(const GenericCtlVec &dx) {
  Log::trace() << "GenericCtlVec::schur_product_with starting" << std::endl;
  util::Timer timer(classname(), "schur_product_with");
  ASSERT(dx.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewDx = atlas::array::make_view<double, 1>(dx.data_);
  for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
    view(jnode) *= viewDx(jnode);
  }
  Log::trace() << "GenericCtlVec::schur_product_with done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::random() {
  Log::trace() << "GenericCtlVec::random starting" << std::endl;
  util::Timer timer(classname(), "random");
  std::default_random_engine generator;
  std::normal_distribution<double> distribution(0.0, 1.0);
  auto view = atlas::array::make_view<double, 1>(data_);
  for (atlas::idx_t jnode = 0; jnode < data_.shape(0); ++jnode) {
    view(jnode) = distribution(generator);
  }
  Log::trace() << "GenericCtlVec::random done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::getSection(std::vector<double> &section, const int &iistart,
                               const int &iiend) const {
  Log::trace() << "GenericCtlVec::getSection starting" << std::endl;
  util::Timer timer(classname(), "getSection");
  ASSERT(iiend >= iistart);
  ASSERT(section.size() > static_cast<size_t>(iiend - iistart));
  auto view = atlas::array::make_view<double, 1>(data_);
  size_t index = 0;
  for (atlas::idx_t jnode = iistart; jnode <= iiend; ++jnode) {
    section[index] = view(jnode);
    ++index;
  }
  Log::trace() << "GenericCtlVec::getSection done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::setSection(const std::vector<double> &section,
                               const int &iistart, const int &iiend) {
  Log::trace() << "GenericCtlVec::setSection starting" << std::endl;
  util::Timer timer(classname(), "setSection");
  ASSERT(iiend >= iistart);
  ASSERT(section.size() > static_cast<size_t>(iiend - iistart));
  auto view = atlas::array::make_view<double, 1>(data_);
  size_t index = 0;
  for (atlas::idx_t jnode = iistart; jnode <= iiend; ++jnode) {
    view(jnode) = section[index];
    ++index;
  }
  Log::trace() << "GenericCtlVec::setSection done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::read(const eckit::Configuration &conf) {
  Log::trace() << "GenericCtlVec::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  ABORT("not implemented yet");
  Log::trace() << "GenericCtlVec::read done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::write(const eckit::Configuration &conf) const {
  Log::trace() << "GenericCtlVec::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  ABORT("not implemented yet");
  Log::trace() << "GenericCtlVec::write done" << std::endl;
}

// -----------------------------------------------------------------------------

double GenericCtlVec::norm() const {
  Log::trace() << "GenericCtlVec::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = this->local_dot_product_with(*this);
  util::mpi::comm().allReduceInPlace(zz, eckit::mpi::sum());
  zz = std::sqrt(zz);
  Log::trace() << "GenericCtlVec::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::print(std::ostream &os) const {
  Log::trace() << "GenericCtlVec::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << "Generic control vector: " << std::endl;
  os << "- name: " << data_.name() << std::endl;
  os << "- size: " << data_.size() << std::endl;
  os << "- norm: " << this->norm() << std::endl;
  Log::trace() << "GenericCtlVec::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops
